;; Generated package description from taxy-magit-section.el  -*- no-byte-compile: t -*-
(define-package "taxy-magit-section" "0.14.2.0.20240919.62537" "View Taxy structs in a Magit Section buffer" '((emacs "26.3") (magit-section "3.2.1") (taxy "0.10")) :commit "b7b60a4be829c28420e3a589ddf3141e4896038b" :authors '(("Adam Porter" . "adam@alphapapa.net")) :maintainer '("Adam Porter" . "adam@alphapapa.net") :keywords '("lisp") :url "https://github.com/alphapapa/taxy.el")
